@@ -12,6 +12,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
+#include <sys/time.h>
 #include <linux/major.h>
 typedef unsigned char u_char;   /* horrible, for scsi.h */
 #include "sg_include.h"
@@ -19,7 +20,7 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
 #include "llseek.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 1999, 2000 D. Gilbert and P. Allworth
+*  Copyright (C) 1999 - 2002 D. Gilbert and P. Allworth
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -40,12 +41,14 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
    in this case) are transferred to or from the sg device in a single SCSI
    command.
 
-   Version 4.12 20010307
+   Version 4.15 20020728
 */
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
 
+#define ME "sg_dd: "
+
 // #define SG_DEBUG
 
 #define DEF_TIMEOUT 40000       /* 40,000 millisecs == 40 seconds */
@@ -65,6 +68,10 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
 #define FT_SG 1                 /* filetype is sg char device */
 #define FT_RAW 2                /* filetype is raw char device */
 
+#define STR_SZ 1024
+#define INOUTF_SZ 512
+#define EBUFF_SZ 512
+
 static unsigned char rdCmdBlk[SCSI_CMD10_LEN] =
                     {0x28, 0, 0, 0, 0, 0, 0, 0, 0, 0};
 static unsigned char wrCmdBlk[SCSI_CMD10_LEN] =
@@ -75,6 +82,7 @@ static int in_full = 0;
 static int in_partial = 0;
 static int out_full = 0;
 static int out_partial = 0;
+static int do_coe = 0;
 
 static void install_handler (int sig_num, void (*sig_handler) (int sig))
 {
@@ -136,9 +144,11 @@ void usage()
 {
     fprintf(stderr, "Usage: "
             "sg_dd  [if=<infile>] [skip=<n>] [of=<ofile>] [seek=<n>]\n"
-            "              [bs=<num>] [bpt=<num>] [count=<n>]\n"
-            "            either 'if' or 'of' must be a sg or raw device\n"
-            " 'bpt' is blocks_per_transfer (default is 128)\n");
+            "              [bs=<num>] [bpt=<num>] [count=<n>] [time=0|1]\n"
+            " where\n"
+            " 'bpt' is blocks_per_transfer (default is 128)\n"
+	    " 'time' 0->no timing(def), 1->time plus calculate throughput\n"
+	    " 'coe' 1->continue on sg error, 0->exit on error (def)\n");
 }
 
 /* Return of 0 -> success, -1 -> failure, 2 -> try again */
@@ -286,7 +296,14 @@ int sg_read(int sg_fd, unsigned char * buff, int blocks, int from_block,
         sg_chk_n_print("reading", osghp->target_status, 
                        osghp->host_status, osghp->driver_status, 
                        osghp->sense_buffer, SG_MAX_SENSE);
-        return -1;
+        if (do_coe) {
+            memset(buff + SG_HEAD_SZ + SCSI_CMD10_LEN, 0, bs * blocks);
+            fprintf(stderr, ">> unable to read at blk=%d for "
+                        "%d bytes, use zeros\n", from_block, bs * blocks);
+            return 0; /* fudge success */
+        }
+        else
+            return -1;
     }
 #else
     if ((osghp->result != 0) || (0 != osghp->sense_buffer[0])) {
@@ -371,7 +388,13 @@ int sg_write(int sg_fd, unsigned char * buff, int blocks, int to_block,
         sg_chk_n_print("writing", osghp->target_status, 
                        osghp->host_status, osghp->driver_status, 
                        osghp->sense_buffer, SG_MAX_SENSE);
-        return -1;
+        if (do_coe) {
+            fprintf(stderr, ">> ignored errors for out blk=%d for "
+                    "%d bytes\n", to_block, bs * blocks);
+            return 0; /* fudge success */
+        }
+        else
+            return -1;
     }
 #else
     if ((osghp->result != 0) || (0 != osghp->sense_buffer[0])) {
@@ -431,13 +454,14 @@ int main(int argc, char * argv[])
     int ibs = 0;
     int obs = 0;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
-    char str[512];
+    char str[STR_SZ];
     char * key;
     char * buf;
-    char inf[512];
+    char inf[INOUTF_SZ];
     int in_type = FT_OTHER;
-    char outf[512];
+    char outf[INOUTF_SZ];
     int out_type = FT_OTHER;
+    int do_time = 0;
     int res, k, t, buf_sz;
     int infd, outfd, blocks;
     unsigned char * wrkBuffRaw;
@@ -446,8 +470,10 @@ int main(int argc, char * argv[])
     int in_num_sect = 0;
     int out_num_sect = 0;
     int in_sect_sz, out_sect_sz;
-    char ebuff[256];
+    char ebuff[EBUFF_SZ];
     int blocks_per;
+    int req_count;
+    struct timeval start_tm, end_tm;
 
     inf[0] = '\0';
     outf[0] = '\0';
@@ -458,7 +484,7 @@ int main(int argc, char * argv[])
 
     for(k = 1; k < argc; k++) {
         if (argv[k])
-            strcpy(str, argv[k]);
+            strncpy(str, argv[k], STR_SZ);
         else
             continue;
         for(key = str, buf = key; *buf && *buf != '=';)
@@ -466,9 +492,9 @@ int main(int argc, char * argv[])
         if (*buf)
             *buf++ = '\0';
         if (strcmp(key,"if") == 0)
-            strcpy(inf, buf);
+            strncpy(inf, buf, INOUTF_SZ);
         else if (strcmp(key,"of") == 0)
-            strcpy(outf, buf);
+            strncpy(outf, buf, INOUTF_SZ);
         else if (0 == strcmp(key,"ibs"))
             ibs = get_num(buf);
         else if (0 == strcmp(key,"obs"))
@@ -483,6 +509,10 @@ int main(int argc, char * argv[])
             seek = get_num(buf);
         else if (0 == strcmp(key,"count"))
             dd_count = get_num(buf);
+        else if (0 == strcmp(key,"coe"))
+            do_coe = get_num(buf);
+        else if (0 == strcmp(key,"time"))
+            do_time = get_num(buf);
         else {
             fprintf(stderr, "Unrecognized argument '%s'\n", key);
             usage();
@@ -503,7 +533,7 @@ int main(int argc, char * argv[])
         return 1;
     }
 #ifdef SG_DEBUG
-    fprintf(stderr, "sg_dd: if=%s skip=%d of=%s seek=%d count=%d\n",
+    fprintf(stderr, ME "if=%s skip=%d of=%s seek=%d count=%d\n",
            inf, skip, outf, seek, dd_count);
 #endif
     install_handler (SIGINT, interrupt_handler);
@@ -518,7 +548,8 @@ int main(int argc, char * argv[])
 
 	if (FT_SG == in_type) {
             if ((infd = open(inf, O_RDWR)) < 0) {
-                sprintf(ebuff, "sg_dd: could not open %s for sg reading", inf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for sg "
+			 "reading", inf);
                 perror(ebuff);
                 return 1;
             }
@@ -526,14 +557,15 @@ int main(int argc, char * argv[])
 #ifdef SG_SET_RESERVED_SIZE
 	    res = ioctl(infd, SG_SET_RESERVED_SIZE, &t);
 	    if (res < 0) {
-		perror("sg_dd: SG_SET_RESERVED_SIZE error");
+		perror(ME "SG_SET_RESERVED_SIZE error");
 		return 1;
 	    }
 #endif
         }
 	if (FT_SG != in_type) {
             if ((infd = open(inf, O_RDONLY)) < 0) {
-                sprintf(ebuff, "sg_dd: could not open %s for reading", inf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for reading",
+			 inf);
                 perror(ebuff);
                 return 1;
             }
@@ -542,8 +574,8 @@ int main(int argc, char * argv[])
 
                 offset *= bs;       /* could exceed 32 bits here! */
                 if (llse_llseek(infd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
-                "sg_dd: couldn't skip to required position on %s", inf);
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't skip to "
+		    	     "required position on %s", inf);
                     perror(ebuff);
                     return 1;
                 }
@@ -555,7 +587,8 @@ int main(int argc, char * argv[])
 
 	if (FT_SG == out_type) {
             if ((outfd = open(outf, O_RDWR)) < 0) {
-                sprintf(ebuff, "sg_dd: could not open %s for sg writing", outf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
+			 "sg writing", outf);
                 perror(ebuff);
                 return 1;
             }
@@ -563,7 +596,7 @@ int main(int argc, char * argv[])
 #ifdef SG_SET_RESERVED_SIZE
 	    res = ioctl(outfd, SG_SET_RESERVED_SIZE, &t);
 	    if (res < 0) {
-		perror("sg_dd: SG_SET_RESERVED_SIZE error");
+		perror(ME "SG_SET_RESERVED_SIZE error");
 		return 1;
 	    }
 #endif
@@ -571,16 +604,16 @@ int main(int argc, char * argv[])
 	else {
 	    if (FT_OTHER == out_type) {
                 if ((outfd = open(outf, O_WRONLY | O_CREAT, 0666)) < 0) {
-                    sprintf(ebuff,
-                            "sg_dd: could not open %s for writing", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "could not open %s "
+		    	     "for writing", outf);
                     perror(ebuff);
                     return 1;
                 }
             }
             else {
                 if ((outfd = open(outf, O_WRONLY)) < 0) {
-                    sprintf(ebuff,
-                            "sg_dd: could not open %s for raw writing", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "could not open %s "
+		    	     "for raw writing", outf);
                     perror(ebuff);
                     return 1;
                 }
@@ -590,8 +623,8 @@ int main(int argc, char * argv[])
 
                 offset *= bs;       /* could overflow here! */
                 if (llse_llseek(outfd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
-                "sg_dd: couldn't seek to required position on %s", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't seek to "
+		    	     "required position on %s", outf);
                     perror(ebuff);
                     return 1;
                 }
@@ -602,12 +635,6 @@ int main(int argc, char * argv[])
         fprintf(stderr, "Can't have both 'if' as stdin _and_ 'of' as stdout\n");
         return 1;
     }
-#if 1
-    if ((FT_OTHER == in_type) && (FT_OTHER == out_type)) {
-        fprintf(stderr, "Either 'if' or 'of' must be a sg or raw device\n");
-        return 1;
-    }
-#endif
     if (0 == dd_count)
         return 0;
     else if (dd_count < 0) {
@@ -697,6 +724,13 @@ int main(int argc, char * argv[])
     fprintf(stderr, "Start of loop, count=%d, blocks_per=%d\n", 
 	    dd_count, blocks_per);
 #endif
+    if (do_time) {
+        start_tm.tv_sec = 0;
+        start_tm.tv_usec = 0;
+        gettimeofday(&start_tm, NULL);
+    }
+    req_count = dd_count;
+
     while (dd_count > 0) {
         blocks = (dd_count > blocks_per) ? blocks_per : dd_count;
         if (FT_SG == in_type) {
@@ -733,7 +767,7 @@ int main(int argc, char * argv[])
                    (EINTR == errno))
                 ;
             if (res < 0) {
-                sprintf(ebuff, "sg_dd: reading, skip=%d ", skip);
+                snprintf(ebuff, EBUFF_SZ, ME "reading, skip=%d ", skip);
                 perror(ebuff);
                 break;
             }
@@ -782,7 +816,7 @@ int main(int argc, char * argv[])
                    && (EINTR == errno))
                 ;
             if (res < 0) {
-                sprintf(ebuff, "sg_ddd512: writing, seek=%d ", seek);
+                snprintf(ebuff, EBUFF_SZ, ME "writing, seek=%d ", seek);
                 perror(ebuff);
                 break;
             }
@@ -802,6 +836,27 @@ int main(int argc, char * argv[])
         skip += blocks;
         seek += blocks;
     }
+    if ((do_time) && (start_tm.tv_sec || start_tm.tv_usec)) {
+        struct timeval res_tm;
+        double a, b;
+
+        gettimeofday(&end_tm, NULL);
+        res_tm.tv_sec = end_tm.tv_sec - start_tm.tv_sec;
+        res_tm.tv_usec = end_tm.tv_usec - start_tm.tv_usec;
+        if (res_tm.tv_usec < 0) {
+            --res_tm.tv_sec;
+            res_tm.tv_usec += 1000000;
+        }
+        a = res_tm.tv_sec;
+        a += (0.000001 * res_tm.tv_usec);
+        b = (double)bs * (req_count - dd_count);
+        printf("time to transfer data was %d.%06d secs",
+               (int)res_tm.tv_sec, (int)res_tm.tv_usec);
+        if ((a > 0.00001) && (b > 511))
+            printf(", %.2f MB/sec\n", b / (a * 1000000.0));
+        else
+            printf("\n");
+    }
 
     free(wrkBuffRaw);
     if (STDIN_FILENO != infd)
