@@ -88,6 +88,11 @@
  * 		sizeof(buffer) (which is sizeof(char*) == 4 or 32 bit archs) 
  *			was used incorrectly all over the place. Fixed.
  *					[version 1.95]
+ * Douglas Gilbert (dgilbert@interlog.com)
+ *    		[skip version 1.96]
+ *    20021211  correct sginfo MODE_SELECT, protect against block devices
+ *              that answer sg's ioctls. [version 1.97]			
+ *    20021228  scan for some "scd<n>" as well as "sr<n>" device names [1.98]
  */
 
 #include <stdio.h>
@@ -99,6 +104,8 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <sys/ioctl.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include "sg_include.h"
 
 #ifdef SG_GET_RESERVED_SIZE
@@ -621,7 +628,7 @@ static int put_mode_page(int page, unsigned char * contents, int page_code)
     pagelen1 = contents[0] + 1;
 
     *((int *) buffer1) = pagelen1+14;      /* length of input data */
-    *(((int *) buffer1) + 1) = pagelen1+8;        /* length of output buffer */
+    *(((int *) buffer1) + 1) = 0;          /* length of output buffer */
 
     cmd = (unsigned char *) (((int *) buffer1) + 2);
 
@@ -1336,15 +1343,18 @@ typedef struct my_scsi_idlun
 
 } My_scsi_idlun;
 
+#define MDEV_NAME_SZ 256
+
 static void make_dev_name(char * fname, int k, int do_numeric)
 {
-    char buff[64];
+    char buff[MDEV_NAME_SZ];
+    size_t len;
 
-    strcpy(fname, "/dev/sg");
-    if (do_numeric) {
-        sprintf(buff, "%d", k);
-        strcat(fname, buff);
-    }
+    strncpy(fname, "/dev/sg", MDEV_NAME_SZ);
+    fname[MDEV_NAME_SZ - 1] = '\0';
+    len = strlen(fname);
+    if (do_numeric)
+    	snprintf(fname + len, MDEV_NAME_SZ - len, "%d", k);
     else {
         if (k <= 26) {
             buff[0] = 'a' + (char)k;
@@ -1364,6 +1374,8 @@ char *devices[] =
  "/dev/sdm", "/dev/sdn", "/dev/sdo", "/dev/sdp", "/dev/sdq", "/dev/sdr",
  "/dev/sds", "/dev/sdt", "/dev/sdu", "/dev/sdv", "/dev/sdw", "/dev/sdx",
  "/dev/sdy", "/dev/sdz", "/dev/sdaa", "/dev/sdab", "/dev/sdac", "/dev/sdad",
+ "/dev/scd0", "/dev/scd1", "/dev/scd2", "/dev/scd3", "/dev/scd4", "/dev/scd5",
+ "/dev/scd6", "/dev/scd7", "/dev/scd8", "/dev/scd9", "/dev/scd10", "/dev/scd11",
  "/dev/sr0", "/dev/sr1", "/dev/sr2", "/dev/sr3", "/dev/sr4", "/dev/sr5",
  "/dev/sr6", "/dev/sr7", "/dev/sr8", "/dev/sr9", "/dev/sr10", "/dev/sr11",
  "/dev/nst0", "/dev/nst1", "/dev/nst2", "/dev/nst3", "/dev/nst4", "/dev/nst5",
@@ -1372,13 +1384,15 @@ char *devices[] =
 
 static Sg_map sg_map_arr[(sizeof(devices) / sizeof(char *)) + 1];
 
+#define EBUFF_SZ 256
+
 /* Print out a list of the known devices on the system */
 static void show_devices()
 {
     int k, j, fd, err, bus;
     My_scsi_idlun m_idlun;
-    char name[64];
-    char ebuff[256];
+    char name[MDEV_NAME_SZ];
+    char ebuff[EBUFF_SZ];
     int do_numeric = 1;
 
     for (k = 0, j = 0; k < sizeof(devices) / sizeof(char *); k++) {
@@ -1387,14 +1401,16 @@ static void show_devices()
             continue;
         err = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &(sg_map_arr[j].bus));
         if (err < 0) {
-            sprintf(ebuff, "SCSI(1) ioctl on %s failed", devices[k]);
+            snprintf(ebuff, EBUFF_SZ,
+	    	     "SCSI(1) ioctl on %s failed", devices[k]);
             perror(ebuff);
             close(fd);
             continue;
         }
         err = ioctl(fd, SCSI_IOCTL_GET_IDLUN, &m_idlun);
         if (err < 0) {
-            sprintf(ebuff, "SCSI(2) ioctl on %s failed", devices[k]);
+            snprintf(ebuff, EBUFF_SZ,
+	    	     "SCSI(2) ioctl on %s failed", devices[k]);
             perror(ebuff);
             close(fd);
             continue;
@@ -1418,7 +1434,7 @@ static void show_devices()
         fd = open(name, O_RDWR | O_NONBLOCK);
         if (fd < 0) {
             if ((ENOENT == errno) && (0 == k)) {
-                do_numeric = 1;
+                do_numeric = 0;
                 make_dev_name(name, k, do_numeric);
                 fd = open(name, O_RDWR | O_NONBLOCK);
             }
@@ -1427,7 +1443,8 @@ static void show_devices()
                     continue;   /* step over if O_EXCL already on it */
                 else {
 #if 0
-                    sprintf(ebuff, "open on %s failed (%d)", name, errno);
+                    snprintf(ebuff, EBUFF_SZ,
+		    	     "open on %s failed (%d)", name, errno);
                     perror(ebuff);
 #endif
                     break;
@@ -1436,14 +1453,14 @@ static void show_devices()
         }
         err = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &bus);
         if (err < 0) {
-            sprintf(ebuff, "SCSI(3) ioctl on %s failed", name);
+            snprintf(ebuff, EBUFF_SZ, "SCSI(3) ioctl on %s failed", name);
             perror(ebuff);
             close(fd);
             continue;
         }
         err = ioctl(fd, SCSI_IOCTL_GET_IDLUN, &m_idlun);
         if (err < 0) {
-            sprintf(ebuff, "SCSI(3) ioctl on %s failed", name);
+            snprintf(ebuff, EBUFF_SZ, "SCSI(3) ioctl on %s failed", name);
             perror(ebuff);
             close(fd);
             continue;
@@ -1637,14 +1654,23 @@ static int open_sg_dev(char * devname)
 {
     int fd, err, bus, bbus, k;
     My_scsi_idlun m_idlun, mm_idlun;
-    int do_numeric = 0;
+    int do_numeric = 1;
     char name[128];
+    struct stat a_st;
+    int block_dev = 0;
 
     strcpy(name, devname);
     fd = open(name, O_RDONLY);
     if (fd < 0)
         return fd;
-    if (ioctl(fd, SG_GET_TIMEOUT, 0) < 0) {
+    if (fstat(fd, &a_st) < 0) {
+	fprintf(stderr, "could do fstat() on fd ??\n");
+	close(fd);
+	return -9999;
+    }
+    if (S_ISBLK(a_st.st_mode))
+	block_dev = 1;
+    if (block_dev || (ioctl(fd, SG_GET_TIMEOUT, 0) < 0)) {
         err = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &bus);
         if (err < 0) {
             perror("A SCSI device name is required\n");
@@ -1664,7 +1690,7 @@ static int open_sg_dev(char * devname)
             fd = open(name, O_RDWR | O_NONBLOCK);
             if (fd < 0) {
                 if ((ENOENT == errno) && (0 == k)) {
-                    do_numeric = 1;
+                    do_numeric = 0;
                     make_dev_name(name, k, do_numeric);
                     fd = open(name, O_RDWR | O_NONBLOCK);
                 }
@@ -1896,7 +1922,7 @@ int main(int argc, char *argv[])
             notch = 1;
             /* fall through */
         case 'v':
-            fprintf(stdout, " Sginfo version 1.95\n");
+            fprintf(stdout, " Sginfo version 1.98\n");
             break;
         default:
             fprintf(stdout, "Unknown option '-%c' (ascii %02xh)\n", c, c);
