@@ -13,6 +13,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
+#include <sys/time.h>
 #include <linux/major.h>
 typedef unsigned char u_char;   /* horrible, for scsi.h */
 #include "sg_include.h"
@@ -20,7 +21,7 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
 #include "llseek.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 1999 - 2001 D. Gilbert and P. Allworth
+*  Copyright (C) 1999 - 2002 D. Gilbert and P. Allworth
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -46,11 +47,13 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
 
 */
 
-static char * version_str = "4.03 20010307";
+static char * version_str = "4.05 20020728";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
 
+#define ME "sgp_dd: "
+
 /* #define SG_DEBUG */
 
 #define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */
@@ -72,6 +75,10 @@ static char * version_str = "4.03 20010307";
 #define FT_SG 1                 /* filetype is sg char device */
 #define FT_RAW 2                /* filetype is raw char device */
 
+#define EBUFF_SZ 512
+#define STR_SZ 1024
+#define INOUTF_SZ 512
+
 
 typedef struct request_collection
 {        /* one instance visible to all threads */
@@ -164,14 +171,14 @@ void usage()
     fprintf(stderr, "Usage: "
            "sgp_dd  [if=<infile>] [skip=<n>] [of=<ofile>] [seek=<n>]\n"
            "               [bs=<num>] [bpt=<num>] [count=<n>]\n"
-           "               [thr=<n>] [coe=<n>] [gen=<n>] [deb=<n>]\n"
-           "               [--version]\n"
-           "            usually either 'if' or 'of' is a sg or raw device\n"
+           "               [thr=<n>] [coe=<n>] [deb=<n>]\n"
+           "               [time=0|1] [--version]\n"
+           " where:\n"
            " 'bpt' is blocks_per_transfer (default is 128)\n"
            " 'thr' is number of threads, must be > 0, default 4, max 16\n");
     fprintf(stderr, " 'coe' continue on error, 0->exit (def), 1->zero "
     	   "+ continue\n"
-           " 'gen' 0-> 1 file is a special(def), 1-> any files allowed\n"
+	   " 'time' 0->no timing(def), 1->time plus calculate throughput\n"
            " 'deb' is debug, 0->none (def), > 0->varying degrees of debug\n");
 }
 
@@ -268,7 +275,7 @@ void * sig_listen_thread(void * v_clp)
     while (1) {
         sigwait(&signal_set, &sig_number);
         if (SIGINT == sig_number) {
-            fprintf(stderr, "sgp_dd interrupted by SIGINT\n");
+            fprintf(stderr, ME "interrupted by SIGINT\n");
 	    guarded_stop_both(clp);
             pthread_cond_broadcast(&clp->out_sync_cv);
         }
@@ -487,7 +494,7 @@ void sg_in_operation(Rq_coll * clp, Rq_elem * rep)
         if (1 == res)
             err_exit(ENOMEM, "sg starting in command");
         else if (res < 0) {
-            fprintf(stderr, "sgp_dd inputting to sg failed, blk=%d\n",
+            fprintf(stderr, ME "inputting to sg failed, blk=%d\n",
                     rep->blk);
             status = pthread_mutex_unlock(&clp->in_mutex);
             if (0 != status) err_exit(status, "unlock in_mutex");
@@ -538,7 +545,7 @@ void sg_out_operation(Rq_coll * clp, Rq_elem * rep)
         if (1 == res)
             err_exit(ENOMEM, "sg starting out command");
         else if (res < 0) {
-            fprintf(stderr, "sgp_dd outputting from sg failed, blk=%d\n",
+            fprintf(stderr, ME "outputting from sg failed, blk=%d\n",
                     rep->blk);
             status = pthread_mutex_unlock(&clp->out_mutex);
             if (0 != status) err_exit(status, "unlock out_mutex");
@@ -647,9 +654,9 @@ int sg_finish_io(int wr, Rq_elem * rep, pthread_mutex_t * a_mutp)
             return 1;
         default:
             {
-                char ebuff[64];
-                sprintf(ebuff, "%s blk=%d", rep->wr ? "writing": "reading",
-                        rep->blk);
+                char ebuff[EBUFF_SZ];
+                snprintf(ebuff, EBUFF_SZ, "%s blk=%d", rep->wr ? 
+					"writing": "reading", rep->blk);
                 status = pthread_mutex_lock(a_mutp);
                 if (0 != status) err_exit(status, "lock aux_mutex");
                 sg_chk_n_print(rep->wr ? "writing": "reading",
@@ -675,27 +682,27 @@ int sg_prepare(int fd, int bs, int bpt, int * scsi_typep)
 
     res = ioctl(fd, SG_GET_VERSION_NUM, &t);
     if ((res < 0) || (t < 20132)) {
-        fprintf(stderr, "sgp_dd: sg driver prior to 2.1.32\n");
+        fprintf(stderr, ME "sg driver prior to 2.1.32\n");
         return 1;
     }
     res = 0;
     t = bs * bpt;
     res = ioctl(fd, SG_SET_RESERVED_SIZE, &t);
     if (res < 0)
-        perror("sgp_dd: SG_SET_RESERVED_SIZE error");
+        perror(ME "SG_SET_RESERVED_SIZE error");
     t = 1;
     res = ioctl(fd, SG_SET_FORCE_PACK_ID, &t);
     if (res < 0)
-        perror("sgp_dd: SG_SET_FORCE_PACK_ID error");
+        perror(ME "SG_SET_FORCE_PACK_ID error");
     res = ioctl(fd, SG_SET_COMMAND_Q, &t);
     if (res < 0)
-        perror("sgp_dd: SG_SET_COMMAND_Q error");
+        perror(ME "SG_SET_COMMAND_Q error");
     if (scsi_typep) {
         struct sg_scsi_id info;
 
         res = ioctl(fd, SG_GET_SCSI_ID, &info);
         if (res < 0)
-            perror("sgp_dd: SG_SET_SCSI_ID error");
+            perror(ME "SG_SET_SCSI_ID error");
         *scsi_typep = info.scsi_type;
     }
     return 0;
@@ -746,20 +753,21 @@ int main(int argc, char * argv[])
     int ibs = 0;
     int obs = 0;
     int count = -1;
-    char str[512];
+    char str[STR_SZ];
     char * key;
     char * buf;
-    char inf[512];
-    char outf[512];
+    char inf[INOUTF_SZ];
+    char outf[INOUTF_SZ];
     int res, k;
     int in_num_sect = 0;
     int out_num_sect = 0;
     int num_threads = DEF_NUM_THREADS;
     pthread_t threads[MAX_NUM_THREADS];
-    int gen = 0;
+    int do_time = 0;
     int in_sect_sz, out_sect_sz, status, infull, outfull;
     void * vp;
-    char ebuff[256];
+    char ebuff[EBUFF_SZ];
+    struct timeval start_tm, end_tm;
     Rq_coll rcoll;
 
     memset(&rcoll, 0, sizeof(Rq_coll));
@@ -775,7 +783,7 @@ int main(int argc, char * argv[])
 
     for(k = 1; k < argc; k++) {
         if (argv[k])
-            strcpy(str, argv[k]);
+            strncpy(str, argv[k], STR_SZ);
         else
             continue;
         for(key = str, buf = key; *buf && *buf != '=';)
@@ -783,9 +791,9 @@ int main(int argc, char * argv[])
         if (*buf)
             *buf++ = '\0';
         if (strcmp(key,"if") == 0)
-            strcpy(inf, buf);
+            strncpy(inf, buf, INOUTF_SZ);
         else if (strcmp(key,"of") == 0)
-            strcpy(outf, buf);
+            strncpy(outf, buf, INOUTF_SZ);
         else if (0 == strcmp(key,"ibs"))
             ibs = get_num(buf);
         else if (0 == strcmp(key,"obs"))
@@ -804,12 +812,12 @@ int main(int argc, char * argv[])
             num_threads = get_num(buf);
         else if (0 == strcmp(key,"coe"))
             rcoll.coe = get_num(buf);
-        else if (0 == strcmp(key,"gen"))
-            gen = get_num(buf);
+        else if (0 == strcmp(key,"time"))
+            do_time = get_num(buf);
         else if (0 == strncmp(key,"deb", 3))
             rcoll.debug = get_num(buf);
         else if (0 == strncmp(key, "--vers", 6)) {
-            fprintf(stderr, "sgp_dd for sg version 2 + 3 driver: %s\n", 
+            fprintf(stderr, ME "for sg version 2 + 3 driver: %s\n", 
 	    	    version_str);
             return 0;
         }
@@ -839,7 +847,7 @@ int main(int argc, char * argv[])
         return 1;
     }
     if (rcoll.debug)
-        fprintf(stderr, "sgp_dd: if=%s skip=%d of=%s seek=%d count=%d\n",
+        fprintf(stderr, ME "if=%s skip=%d of=%s seek=%d count=%d\n",
                inf, skip, outf, seek, count);
     rcoll.infd = STDIN_FILENO;
     rcoll.outfd = STDOUT_FILENO;
@@ -848,8 +856,8 @@ int main(int argc, char * argv[])
 
         if (FT_SG == rcoll.in_type) {
             if ((rcoll.infd = open(inf, O_RDWR)) < 0) {
-                sprintf(ebuff, "sgp_dd: could not open %s for sg reading",
-                        inf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
+			 "sg reading", inf);
                 perror(ebuff);
                 return 1;
             }
@@ -859,7 +867,8 @@ int main(int argc, char * argv[])
         }
         if (FT_SG != rcoll.in_type) {
             if ((rcoll.infd = open(inf, O_RDONLY)) < 0) {
-                sprintf(ebuff, "sgp_dd: could not open %s for reading", inf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for reading",
+			 inf);
                 perror(ebuff);
                 return 1;
             }
@@ -868,8 +877,8 @@ int main(int argc, char * argv[])
 
                 offset *= rcoll.bs;       /* could exceed 32 here! */
                 if (llse_llseek(rcoll.infd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
-                "sgp_dd: couldn't skip to required position on %s", inf);
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't skip to required"
+			     " position on %s", inf);
                     perror(ebuff);
                     return 1;
                 }
@@ -881,8 +890,8 @@ int main(int argc, char * argv[])
 
         if (FT_SG == rcoll.out_type) {
             if ((rcoll.outfd = open(outf, O_RDWR)) < 0) {
-                sprintf(ebuff,
-                        "sgp_dd: could not open %s for sg writing", outf);
+                snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
+			 "sg writing", outf);
                 perror(ebuff);
                 return 1;
             }
@@ -894,16 +903,16 @@ int main(int argc, char * argv[])
         else {
             if (FT_OTHER == rcoll.out_type) {
                 if ((rcoll.outfd = open(outf, O_WRONLY | O_CREAT, 0666)) < 0) {
-                    sprintf(ebuff,
-                            "sgp_dd: could not open %s for writing", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "could not open %s "
+			     "for writing", outf);
                     perror(ebuff);
                     return 1;
                 }
             }
             else {
                 if ((rcoll.outfd = open(outf, O_WRONLY)) < 0) {
-                    sprintf(ebuff,
-                            "sgp_dd: could not open %s for raw writing", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
+			     "raw writing", outf);
                     perror(ebuff);
                     return 1;
                 }
@@ -913,8 +922,8 @@ int main(int argc, char * argv[])
 
                 offset *= rcoll.bs;       /* could exceed 32 bits here! */
                 if (llse_llseek(rcoll.outfd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
-                "sgp_dd: couldn't seek to required position on %s", outf);
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't seek to "
+			     "required position on %s", outf);
                     perror(ebuff);
                     return 1;
                 }
@@ -925,10 +934,6 @@ int main(int argc, char * argv[])
         fprintf(stderr, "Disallow both if and of to be stdin and stdout");
         return 1;
     }
-    if ((FT_OTHER == rcoll.in_type) && (FT_OTHER == rcoll.out_type) && !gen) {
-        fprintf(stderr, "Either 'if' or 'of' must be a sg or raw device\n");
-        return 1;
-    }
     if (0 == count)
         return 0;
     else if (count < 0) {
@@ -1005,6 +1010,12 @@ int main(int argc, char * argv[])
                             sig_listen_thread, (void *)&rcoll);
     if (0 != status) err_exit(status, "pthread_create, sig...");
 
+    if (do_time) {
+        start_tm.tv_sec = 0;
+        start_tm.tv_usec = 0;
+        gettimeofday(&start_tm, NULL);
+    }
+
 /* vvvvvvvvvvv  Start worker threads  vvvvvvvvvvvvvvvvvvvvvvvv */
     if ((rcoll.out_done_count > 0) && (num_threads > 0)) {
         /* Run 1 work thread to shake down infant retryable stuff */
@@ -1042,6 +1053,28 @@ int main(int argc, char * argv[])
         }
     }
 
+    if ((do_time) && (start_tm.tv_sec || start_tm.tv_usec)) {
+        struct timeval res_tm;
+        double a, b;
+
+        gettimeofday(&end_tm, NULL);
+        res_tm.tv_sec = end_tm.tv_sec - start_tm.tv_sec;
+        res_tm.tv_usec = end_tm.tv_usec - start_tm.tv_usec;
+        if (res_tm.tv_usec < 0) {
+            --res_tm.tv_sec;
+            res_tm.tv_usec += 1000000;
+        }
+        a = res_tm.tv_sec;
+        a += (0.000001 * res_tm.tv_usec);
+        b = (double)rcoll.bs * (count - rcoll.out_done_count);
+        printf("time to transfer data was %d.%06d secs",
+               (int)res_tm.tv_sec, (int)res_tm.tv_usec);
+        if ((a > 0.00001) && (b > 511))
+            printf(", %.2f MB/sec\n", b / (a * 1000000.0));
+        else
+            printf("\n");
+    }
+
     status = pthread_cancel(sig_listen_thread_id);
     if (0 != status) err_exit(status, "pthread_cancel");
     if (STDIN_FILENO != rcoll.infd)
